
/******************************************************************************
* MODULE     : symbol.cc
* DESCRIPTION: Symbols
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <symbol.hh>
namespace mmx {

GC_GLOBAL (table, symbol_table, table ()); // GL: hack for completion

/******************************************************************************
* Required functions for generic structures
******************************************************************************/

size_t
symbol_rep::Size () {
  return sizeof (symbol_rep);
}

void
symbol_rep::Mark () {
  mark (hash_value);
  mark (ref);
}

int
symbol_rep::Type () {
  return SYMBOL_TYPE;
}

int
symbol_rep::Hash () {
  return hash_value;
}

bool
symbol_rep::Equal (generic g) {
  if (gentype (g) != SYMBOL_TYPE) return false;
  return symbol (this) == as_symbol (g);
}

generic
symbol_rep::Copy () {
  return copy (symbol (this));
}

void
symbol_rep::Print (ostream& out) {
  out << symbol (this);
}

/******************************************************************************
* Constructors and convertors for symbols
******************************************************************************/

inline
symbol_rep::symbol_rep (generic g):
  hash_value (hash(g)^0x55555555)
{
  if (!symbol_table->contains (g))
    symbol_table (g)= g;
  ref= symbol_table[g];
}

symbol::symbol (char* s) {
  generic g= string (s);
  rep= new symbol_rep (g);
}

symbol::symbol (string s) {
  generic g= string (s);
  rep= new symbol_rep (g);
}

symbol::symbol (generic g) {
  rep= new symbol_rep (g);
}

symbol::symbol (char* s, int begin, int end) {
  generic g= string (s);
  rep= new symbol_rep (g);
  rep -> index (begin, end);
}

symbol::symbol (string s, int begin, int end) {
  generic g= s;
  rep= new symbol_rep (g);
  rep -> index (begin, end);
}

symbol::symbol (generic g, int begin, int end) {
  rep= new symbol_rep (g);
  rep -> index (begin, end);
}

symbol::symbol (char* s, generic where) {
  generic g= string (s);
  rep= new symbol_rep (g);
  rep -> index (where);
}

symbol::symbol (string s, generic where) {
  generic g= s;
  rep= new symbol_rep (g);
  rep -> index (where);
}

symbol::symbol (generic g, generic where) {
  rep= new symbol_rep (g);
  rep -> index (where);
}

ostream&
operator << (ostream& out, symbol s) {
  generic g= S (s);
  if (gentype (g) == STRING_TYPE && !long_print_format ())
    out << as_string (g);
  else if (current_print_format == indented) {
    out << "#symbol";
    current_indentation++;
    out << lf << g;
    current_indentation--;
  }
  else out << "#symbol<" << g << ">";
  return out;
}

/******************************************************************************
* Other converters
******************************************************************************/

bool
is_int (symbol x) {
  generic g= S(x);
  if (gentype (g) != STRING_TYPE) return false;
  string s= as_string (g);
  int i=0, n=N(s);
  if ((i<n) && ((s[i]=='+') || (s[i]=='-'))) i++;
  if (i>=n) return false;
  for (; i<n; i++)
    if ((s[i]<'0') || (s[i]>'9')) return false;
  return true;
}

bool
is_double (symbol x) {
  generic g= S(x);
  if (gentype (g) != STRING_TYPE) return false;
  string s= as_string (g);
  int i=0, n=N(s);
  if ((i<n) && ((s[i]=='+') || (s[i]=='-'))) i++;
  if (i>=n) return false;
  for (; i<n; i++)
    if (s[i]=='.') break;
    else if ((s[i]<'0') || (s[i]>'9')) return false;
  if (i>=n) return false;
  for (i++; i<n; i++)
    if ((s[i]=='e') || (s[i]=='E')) break;
    else if ((s[i]<'0') || (s[i]>'9')) return false;
  if (i>=n) return true;
  i++;
  if ((i<n) && ((s[i]=='+') || (s[i]=='-'))) i++;
  if (i>=n) return false;
  for (; i<n; i++)
    if ((s[i]<'0') || (s[i]>'9')) return false;
  return true;
}


bool
is_string (symbol x) {
  generic g= S(x);
  if (gentype (g) != STRING_TYPE) return false;
  string s= as_string (g);
  int n=N(s);
  return (n>=2) && (s[0]=='\042') && (s[n-1]=='\042');
}

int
as_int (symbol x) {
  generic g= S(x);
  if (gentype (g) != STRING_TYPE) return 0;
  string s= as_string (g);
  int i=0, n=N(s), val=0;
  if (n==0) return 0;
  if ((s[0]=='+') || (s[0]=='-')) i++;
  while (i<n) {
    if (s[i]<'0') break;
    if (s[i]>'9') break;
    val *= 10;
    val += (int) (s[i]-'0');
    i++;
  }
  if (s[0]=='-') val=-val;
  return val;
}

double
as_double (symbol x) {
  generic g= S(x);
  if (gentype (g) != STRING_TYPE) return 0.0;
  return as_double (as_string (g));
}

string 
as_string (symbol x) {
  generic g= S(x);
  if (gentype (g) != STRING_TYPE) return "";
  string s= as_string (g);
  if (N(s)<2) return "";
  return s (1, N(s)-1);
}

} // namespace mmx
